@@ -1175,6 +1175,17 @@ make_spec_file(void)
 	printf("netdump, diskdump and kdump facilities from Red Hat Linux, the mcore kernel patch\n");
 	printf("offered by Mission Critical Linux, or the LKCD kernel patch.\n");
 	printf("\n");
+	printf("%%package devel\n");
+	printf("Requires: %%{name} = %%{version}\n");
+	printf("Summary: crash utility for live systems; netdump, diskdump, kdump, LKCD or mcore dumpfiles\n");
+	printf("Group: Development/Debuggers\n");
+	printf("\n");
+	printf("%%description devel\n");
+	printf("The core analysis suite is a self-contained tool that can be used to\n");
+	printf("investigate either live systems, kernel core dumps created from the\n");
+	printf("netdump, diskdump and kdump packages from Red Hat Linux, the mcore kernel patch\n");
+	printf("offered by Mission Critical Linux, or the LKCD kernel patch.\n");
+	printf("\n");
 	printf("%%prep\n");
         printf("%%setup -n %%{name}-%%{version}-%%{release}\n"); 
 	printf("# %%patch0 -p1 -b .install (patch example)\n");
@@ -1189,6 +1200,8 @@ make_spec_file(void)
 	printf("make DESTDIR=%%{buildroot} install\n");
 	printf("mkdir -p %%{buildroot}%%{_mandir}/man8\n");
 	printf("cp crash.8 %%{buildroot}%%{_mandir}/man8/crash.8\n");
+	printf("mkdir -p %%{buildroot}%%{_includedir}/crash\n");
+	printf("cp defs.h %%{buildroot}%%{_includedir}/crash\n");
 	printf("\n");
 	printf("%%clean\n");
 	printf("rm -rf %%{buildroot}\n");
@@ -1198,6 +1211,10 @@ make_spec_file(void)
 	printf("%%{_mandir}/man8/crash.8*\n");
      /*	printf("/usr/bin/crashd\n"); */
 	printf("%%doc README\n");
+	printf("\n");
+	printf("%%files devel\n");
+	printf("%%defattr(-,root,root)\n");
+	printf("%%{_includedir}/*\n");
 }
 
 /*
