@@ -1,6 +1,8 @@
 /* Support for GDB maintenance commands.
-   Copyright 1992, 1993, 1994, 1995, 1996, 1997, 1999, 2000, 2001
-   Free Software Foundation, Inc.
+
+   Copyright 1992, 1993, 1994, 1995, 1996, 1997, 1999, 2000, 2001,
+   2002, 2003, 2004 Free Software Foundation, Inc.
+
    Written by Fred Fish at Cygnus Support.
 
    This file is part of GDB.
@@ -36,6 +38,8 @@
 #include "objfiles.h"
 #include "value.h"
 
+#include "cli/cli-decode.h"
+
 extern void _initialize_maint_cmds (void);
 
 static void maintenance_command (char *, int);
@@ -52,8 +56,6 @@ static void maintenance_space_display (char *, int);
 
 static void maintenance_info_command (char *, int);
 
-static void print_section_table (bfd *, asection *, PTR);
-
 static void maintenance_info_sections (char *, int);
 
 static void maintenance_print_command (char *, int);
@@ -90,7 +92,6 @@ maintenance_command (char *args, int from_tty)
 }
 
 #ifndef _WIN32
-/* ARGSUSED */
 static void
 maintenance_dump_me (char *args, int from_tty)
 {
@@ -116,8 +117,18 @@ maintenance_dump_me (char *args, int from_tty)
 static void
 maintenance_internal_error (char *args, int from_tty)
 {
-  internal_error (__FILE__, __LINE__,
-		  "internal maintenance");
+  internal_error (__FILE__, __LINE__, "%s", (args == NULL ? "" : args));
+}
+
+/* Stimulate the internal error mechanism that GDB uses when an
+   internal problem is detected.  Allows testing of the mechanism.
+   Also useful when the user wants to drop a core file but not exit
+   GDB. */
+
+static void
+maintenance_internal_warning (char *args, int from_tty)
+{
+  internal_warning (__FILE__, __LINE__, "%s", (args == NULL ? "" : args));
 }
 
 /* Someday we should allow demangling for things other than just
@@ -139,7 +150,8 @@ maintenance_demangle (char *args, int from_tty)
     }
   else
     {
-      demangled = cplus_demangle (args, DMGL_ANSI | DMGL_PARAMS);
+      demangled = language_demangle (current_language, args, 
+				     DMGL_ANSI | DMGL_PARAMS);
       if (demangled != NULL)
 	{
 	  printf_unfiltered ("%s\n", demangled);
@@ -178,7 +190,6 @@ maintenance_space_display (char *args, int from_tty)
    allow_unknown 0.  Therefore, its own definition is called only for
    "maintenance info" with no args.  */
 
-/* ARGSUSED */
 static void
 maintenance_info_command (char *arg, int from_tty)
 {
@@ -186,27 +197,82 @@ maintenance_info_command (char *arg, int from_tty)
   help_list (maintenanceinfolist, "maintenance info ", -1, gdb_stdout);
 }
 
-static void
-print_section_table (bfd *abfd, asection *asect, PTR ignore)
+/* Mini tokenizing lexer for 'maint info sections' command.  */
+
+static int
+match_substring (const char *string, const char *substr)
 {
-  flagword flags;
+  int substr_len = strlen(substr);
+  const char *tok;
 
-  flags = bfd_get_section_flags (abfd, asect);
+  while ((tok = strstr (string, substr)) != NULL)
+    {
+      /* Got a partial match.  Is it a whole word? */
+      if (tok == string
+	  || tok[-1] == ' '
+	  || tok[-1] == '\t')
+      {
+	/* Token is delimited at the front... */
+	if (tok[substr_len] == ' '
+	    || tok[substr_len] == '\t'
+	    || tok[substr_len] == '\0')
+	{
+	  /* Token is delimited at the rear.  Got a whole-word match.  */
+	  return 1;
+	}
+      }
+      /* Token didn't match as a whole word.  Advance and try again.  */
+      string = tok + 1;
+    }
+  return 0;
+}
 
-  /* FIXME-32x64: Need print_address_numeric with field width.  */
-  printf_filtered ("    %s",
-		   local_hex_string_custom
-		   ((unsigned long) bfd_section_vma (abfd, asect), "08l"));
-  printf_filtered ("->%s",
-		   local_hex_string_custom
-		   ((unsigned long) (bfd_section_vma (abfd, asect)
-				     + bfd_section_size (abfd, asect)),
-		    "08l"));
-  printf_filtered (" at %s",
-		   local_hex_string_custom
-		   ((unsigned long) asect->filepos, "08l"));
-  printf_filtered (": %s", bfd_section_name (abfd, asect));
+static int 
+match_bfd_flags (char *string, flagword flags)
+{
+  if (flags & SEC_ALLOC)
+    if (match_substring (string, "ALLOC"))
+      return 1;
+  if (flags & SEC_LOAD)
+    if (match_substring (string, "LOAD"))
+      return 1;
+  if (flags & SEC_RELOC)
+    if (match_substring (string, "RELOC"))
+      return 1;
+  if (flags & SEC_READONLY)
+    if (match_substring (string, "READONLY"))
+      return 1;
+  if (flags & SEC_CODE)
+    if (match_substring (string, "CODE"))
+      return 1;
+  if (flags & SEC_DATA)
+    if (match_substring (string, "DATA"))
+      return 1;
+  if (flags & SEC_ROM)
+    if (match_substring (string, "ROM"))
+      return 1;
+  if (flags & SEC_CONSTRUCTOR)
+    if (match_substring (string, "CONSTRUCTOR"))
+      return 1;
+  if (flags & SEC_HAS_CONTENTS)
+    if (match_substring (string, "HAS_CONTENTS"))
+      return 1;
+  if (flags & SEC_NEVER_LOAD)
+    if (match_substring (string, "NEVER_LOAD"))
+      return 1;
+  if (flags & SEC_COFF_SHARED_LIBRARY)
+    if (match_substring (string, "COFF_SHARED_LIBRARY"))
+      return 1;
+  if (flags & SEC_IS_COMMON)
+    if (match_substring (string, "IS_COMMON"))
+      return 1;
 
+  return 0;
+}
+
+static void
+print_bfd_flags (flagword flags)
+{
   if (flags & SEC_ALLOC)
     printf_filtered (" ALLOC");
   if (flags & SEC_LOAD)
@@ -231,11 +297,60 @@ print_section_table (bfd *abfd, asection *asect, PTR ignore)
     printf_filtered (" COFF_SHARED_LIBRARY");
   if (flags & SEC_IS_COMMON)
     printf_filtered (" IS_COMMON");
+}
 
+static void
+maint_print_section_info (const char *name, flagword flags, 
+			  CORE_ADDR addr, CORE_ADDR endaddr, 
+			  unsigned long filepos)
+{
+  /* FIXME-32x64: Need print_address_numeric with field width.  */
+  printf_filtered ("    0x%s", paddr (addr));
+  printf_filtered ("->0x%s", paddr (endaddr));
+  printf_filtered (" at %s",
+		   local_hex_string_custom ((unsigned long) filepos, "08l"));
+  printf_filtered (": %s", name);
+  print_bfd_flags (flags);
   printf_filtered ("\n");
 }
 
-/* ARGSUSED */
+static void
+print_bfd_section_info (bfd *abfd, 
+			asection *asect, 
+			void *arg)
+{
+  flagword flags = bfd_get_section_flags (abfd, asect);
+  const char *name = bfd_section_name (abfd, asect);
+
+  if (arg == NULL || *((char *) arg) == '\0'
+      || match_substring ((char *) arg, name)
+      || match_bfd_flags ((char *) arg, flags))
+    {
+      CORE_ADDR addr, endaddr;
+
+      addr = bfd_section_vma (abfd, asect);
+      endaddr = addr + bfd_section_size (abfd, asect);
+      maint_print_section_info (name, flags, addr, endaddr, asect->filepos);
+    }
+}
+
+static void
+print_objfile_section_info (bfd *abfd, 
+			    struct obj_section *asect, 
+			    char *string)
+{
+  flagword flags = bfd_get_section_flags (abfd, asect->the_bfd_section);
+  const char *name = bfd_section_name (abfd, asect->the_bfd_section);
+
+  if (string == NULL || *string == '\0'
+      || match_substring (string, name)
+      || match_bfd_flags (string, flags))
+    {
+      maint_print_section_info (name, flags, asect->addr, asect->endaddr, 
+			  asect->the_bfd_section->filepos);
+    }
+}
+
 static void
 maintenance_info_sections (char *arg, int from_tty)
 {
@@ -245,7 +360,30 @@ maintenance_info_sections (char *arg, int from_tty)
       printf_filtered ("    `%s', ", bfd_get_filename (exec_bfd));
       wrap_here ("        ");
       printf_filtered ("file type %s.\n", bfd_get_target (exec_bfd));
-      bfd_map_over_sections (exec_bfd, print_section_table, 0);
+      if (arg && *arg && match_substring (arg, "ALLOBJ"))
+	{
+	  struct objfile *ofile;
+	  struct obj_section *osect;
+
+	  /* Only this function cares about the 'ALLOBJ' argument; 
+	     if 'ALLOBJ' is the only argument, discard it rather than
+	     passing it down to print_objfile_section_info (which 
+	     wouldn't know how to handle it).  */
+	  if (strcmp (arg, "ALLOBJ") == 0)
+	    arg = NULL;
+
+	  ALL_OBJFILES (ofile)
+	    {
+	      printf_filtered ("  Object file: %s\n", 
+			       bfd_get_filename (ofile->obfd));
+	      ALL_OBJFILE_OSECTIONS (ofile, osect)
+		{
+		  print_objfile_section_info (ofile->obfd, osect, arg);
+		}
+	    }
+	}
+      else 
+	bfd_map_over_sections (exec_bfd, print_bfd_section_info, arg);
     }
 
   if (core_bfd)
@@ -254,11 +392,10 @@ maintenance_info_sections (char *arg, int from_tty)
       printf_filtered ("    `%s', ", bfd_get_filename (core_bfd));
       wrap_here ("        ");
       printf_filtered ("file type %s.\n", bfd_get_target (core_bfd));
-      bfd_map_over_sections (core_bfd, print_section_table, 0);
+      bfd_map_over_sections (core_bfd, print_bfd_section_info, arg);
     }
 }
 
-/* ARGSUSED */
 void
 maintenance_print_statistics (char *args, int from_tty)
 {
@@ -266,7 +403,7 @@ maintenance_print_statistics (char *args, int from_tty)
   print_symbol_bcache_statistics ();
 }
 
-void
+static void
 maintenance_print_architecture (char *args, int from_tty)
 {
   if (args == NULL)
@@ -285,7 +422,6 @@ maintenance_print_architecture (char *args, int from_tty)
    allow_unknown 0.  Therefore, its own definition is called only for
    "maintenance print" with no args.  */
 
-/* ARGSUSED */
 static void
 maintenance_print_command (char *arg, int from_tty)
 {
@@ -344,7 +480,7 @@ maintenance_translate_address (char *arg, int from_tty)
 
   if (sym)
     printf_filtered ("%s+%s\n",
-		     SYMBOL_SOURCE_NAME (sym),
+		     SYMBOL_PRINT_NAME (sym),
 		     paddr_u (address - SYMBOL_VALUE_ADDRESS (sym)));
   else if (sect)
     printf_filtered ("no symbol at %s:0x%s\n", sect->name, paddr (address));
@@ -471,15 +607,98 @@ maintenance_do_deprecate (char *text, int deprecate)
     }
 }
 
+/* Maintenance set/show framework.  */
+
+static struct cmd_list_element *maintenance_set_cmdlist;
+static struct cmd_list_element *maintenance_show_cmdlist;
+
+static void
+maintenance_set_cmd (char *args, int from_tty)
+{
+  printf_unfiltered ("\"maintenance set\" must be followed by the name of a set command.\n");
+  help_list (maintenance_set_cmdlist, "maintenance set ", -1, gdb_stdout);
+}
+
+static void
+maintenance_show_cmd (char *args, int from_tty)
+{
+  cmd_show_list (maintenance_show_cmdlist, from_tty, "");
+}
+
+/* Profiling support.  */
+
+static int maintenance_profile_p;
+
+#if defined (HAVE_MONSTARTUP) && defined (HAVE__MCLEANUP)
+
+#ifdef HAVE__ETEXT
+extern char _etext;
+#define TEXTEND &_etext
+#else
+extern char etext;
+#define TEXTEND &etext
+#endif
+
+static int profiling_state;
+
+static void
+mcleanup_wrapper (void)
+{
+  extern void _mcleanup (void);
+
+  if (profiling_state)
+    _mcleanup ();
+}
+
+static void
+maintenance_set_profile_cmd (char *args, int from_tty, struct cmd_list_element *c)
+{
+  if (maintenance_profile_p == profiling_state)
+    return;
+
+  profiling_state = maintenance_profile_p;
+
+  if (maintenance_profile_p)
+    {
+      static int profiling_initialized;
+
+      extern void monstartup (unsigned long, unsigned long);
+      extern int main();
+
+      if (!profiling_initialized)
+	{
+	  atexit (mcleanup_wrapper);
+	  profiling_initialized = 1;
+	}
+
+      /* "main" is now always the first function in the text segment, so use
+	 its address for monstartup.  */
+      monstartup ((unsigned long) &main, (unsigned long) TEXTEND);
+    }
+  else
+    {
+      extern void _mcleanup (void);
+      _mcleanup ();
+    }
+}
+#else
+static void
+maintenance_set_profile_cmd (char *args, int from_tty, struct cmd_list_element *c)
+{
+  error ("Profiling support is not available on this system.");
+}
+#endif
 
 void
 _initialize_maint_cmds (void)
 {
+  struct cmd_list_element *tmpcmd;
+
   add_prefix_cmd ("maintenance", class_maintenance, maintenance_command,
 		  "Commands for use by GDB maintainers.\n\
 Includes commands to dump specific internal GDB structures in\n\
 a human readable form, to cause GDB to deliberately dump core,\n\
-to test internal functions such as the C++ demangler, etc.",
+to test internal functions such as the C++/ObjC demangler, etc.",
 		  &maintenancelist, "maintenance ", 0,
 		  &cmdlist);
 
@@ -492,7 +711,15 @@ to test internal functions such as the C++ demangler, etc.",
   add_alias_cmd ("i", "info", class_maintenance, 1, &maintenancelist);
 
   add_cmd ("sections", class_maintenance, maintenance_info_sections,
-	   "List the BFD sections of the exec and core files.",
+	   "List the BFD sections of the exec and core files. \n\
+Arguments may be any combination of:\n\
+	[one or more section names]\n\
+	ALLOC LOAD RELOC READONLY CODE DATA ROM CONSTRUCTOR\n\
+	HAS_CONTENTS NEVER_LOAD COFF_SHARED_LIBRARY IS_COMMON\n\
+Sections matching any argument will be listed (no argument\n\
+implies all sections).  In addition, the special argument\n\
+	ALLOBJ\n\
+lists all sections from all object files, including shared libraries.",
 	   &maintenanceinfolist);
 
   add_prefix_cmd ("print", class_maintenance, maintenance_print_command,
@@ -500,10 +727,24 @@ to test internal functions such as the C++ demangler, etc.",
 		  &maintenanceprintlist, "maintenance print ", 0,
 		  &maintenancelist);
 
+  add_prefix_cmd ("set", class_maintenance, maintenance_set_cmd, "\
+Set GDB internal variables used by the GDB maintainer.\n\
+Configure variables internal to GDB that aid in GDB's maintenance",
+		  &maintenance_set_cmdlist, "maintenance set ",
+		  0/*allow-unknown*/,
+		  &maintenancelist);
+
+  add_prefix_cmd ("show", class_maintenance, maintenance_show_cmd, "\
+Show GDB internal variables used by the GDB maintainer.\n\
+Configure variables internal to GDB that aid in GDB's maintenance",
+		  &maintenance_show_cmdlist, "maintenance show ",
+		  0/*allow-unknown*/,
+		  &maintenancelist);
+
 #ifndef _WIN32
   add_cmd ("dump-me", class_maintenance, maintenance_dump_me,
 	   "Get fatal error; make debugger dump its core.\n\
-GDB sets it's handling of SIGQUIT back to SIG_DFL and then sends\n\
+GDB sets its handling of SIGQUIT back to SIG_DFL and then sends\n\
 itself a SIGQUIT signal.",
 	   &maintenancelist);
 #endif
@@ -513,8 +754,13 @@ itself a SIGQUIT signal.",
 Cause GDB to behave as if an internal error was detected.",
 	   &maintenancelist);
 
+  add_cmd ("internal-warning", class_maintenance, maintenance_internal_warning,
+	   "Give GDB an internal warning.\n\
+Cause GDB to behave as if an internal warning was reported.",
+	   &maintenancelist);
+
   add_cmd ("demangle", class_maintenance, maintenance_demangle,
-	   "Demangle a C++ mangled name.\n\
+	   "Demangle a C++/ObjC mangled name.\n\
 Call internal GDB demangler routine to demangle a C++ link name\n\
 and prints the result.",
 	   &maintenancelist);
@@ -559,6 +805,19 @@ If a SOURCE file is specified, dump only that file's partial symbols.",
 	   "Print dump of current object file definitions.",
 	   &maintenanceprintlist);
 
+  add_cmd ("symtabs", class_maintenance, maintenance_info_symtabs,
+	   "List the full symbol tables for all object files.\n\
+This does not include information about individual symbols, blocks, or\n\
+linetables --- just the symbol table structures themselves.\n\
+With an argument REGEXP, list the symbol tables whose names that match that.",
+	   &maintenanceinfolist);
+
+  add_cmd ("psymtabs", class_maintenance, maintenance_info_psymtabs,
+	   "List the partial symbol tables for all object files.\n\
+This does not include information about individual partial symbols,\n\
+just the symbol table structures themselves.",
+	   &maintenanceinfolist);
+
   add_cmd ("statistics", class_maintenance, maintenance_print_statistics,
 	   "Print statistics about internal gdb state.",
 	   &maintenanceprintlist);
@@ -596,4 +855,14 @@ When non-zero, this timeout is used instead of waiting forever for a target to\n
 finish a low-level step or continue operation.  If the specified amount of time\n\
 passes without a response from the target, an error occurs.", &setlist),
 		      &showlist);
+
+
+  add_setshow_boolean_cmd ("profile", class_maintenance,
+			   &maintenance_profile_p,
+			   "Set internal profiling.\n"
+			   "When enabled GDB is profiled.",
+			   "Show internal profiling.\n",
+			   maintenance_set_profile_cmd, NULL,
+			   &maintenance_set_cmdlist,
+			   &maintenance_show_cmdlist);
 }
