@@ -74,14 +74,13 @@
 #define PERCPU_ADDR	      (DEFAULT_SHAREDINFO_ADDR - PERCPU_PAGE_SIZE)
 #define DIRECTMAP_VIRT_START  (0xf000000000000000)
 #define DIRECTMAP_VIRT_END    PERCPU_ADDR
-#define VIRT_FRAME_TABLE_ADDR    (0xf300000000000000)
-#define VIRT_FRAME_TABLE_END     (0xf400000000000000)
+#define VIRT_FRAME_TABLE_SIZE     (0x0100000000000000)
 
 #define PERCPU_VIRT_ADDR(vaddr) \
     (((vaddr) >= PERCPU_ADDR) && ((vaddr) < PERCPU_ADDR + PERCPU_PAGE_SIZE))
 
 #define FRAME_TABLE_VIRT_ADDR(vaddr) \
-    ((vaddr) >= VIRT_FRAME_TABLE_ADDR && (vaddr) < VIRT_FRAME_TABLE_END)
+    ((vaddr) >= xhmachdep->frame_table && (vaddr) < xhmachdep->frame_table + VIRT_FRAME_TABLE_SIZE)
 
 #undef IA64_RBS_OFFSET
 #define IA64_RBS_OFFSET   ((XEN_HYPER_SIZE(vcpu) + 15) & ~15)
@@ -101,6 +100,14 @@ typedef uint32_t	Elf_Word;
 #define XEN_HYPER_ELF_NOTE_V1	1
 #define XEN_HYPER_ELF_NOTE_V2	2
 #define XEN_HYPER_ELF_NOTE_V3	3
+#define XEN_HYPER_ELF_NOTE_V4	4
+
+#ifdef X86
+#define XEN_HYPER_ELF_NOTE_V4_NOTE_SIZE	0x100
+#endif
+#if defined(X86_64) || defined(IA64)
+#define XEN_HYPER_ELF_NOTE_V4_NOTE_SIZE	0x200
+#endif
 
 /*
  * Xen Hyper
@@ -129,7 +136,9 @@ typedef uint32_t	Elf_Word;
 	((ulong)(var) + (((ulong)(cpu))<<XEN_HYPER_PERCPU_SHIFT))
 #elif defined(IA64)
 #define xen_hyper_per_cpu(var, cpu)  \
-	((ulong)(var) + (xht->__per_cpu_offset[cpu]))
+	((xht->flags & XEN_HYPER_SMP) ? \
+		(ulong)(var) + (xht->__per_cpu_offset[cpu]) : \
+		(ulong)(var))
 #endif
 
 #if defined(X86) || defined(X86_64)
@@ -155,6 +164,7 @@ typedef uint32_t	Elf_Word;
 /* Prepared domain ID. */
 #define XEN_HYPER_DOMID_IO		(0x7FF1U)
 #define XEN_HYPER_DOMID_XEN		(0x7FF2U)
+
 /* Domain flags (domain_flags). */
  /* Is this domain privileged? */
 #define XEN_HYPER__DOMF_privileged       0
@@ -180,6 +190,32 @@ typedef uint32_t	Elf_Word;
  /* Domain flag error */
 #define XEN_HYPER_DOMF_ERROR            ((ulong)(-1))
 
+/* Domain status. */
+ /* Is this an HVM guest? */
+#define XEN_HYPER__DOMS_HVM              0
+#define XEN_HYPER_DOMS_HVM               (1UL<<XEN_HYPER__DOMS_HVM)
+ /* Is this guest fully privileged (aka dom0)? */
+#define XEN_HYPER__DOMS_privileged       1
+#define XEN_HYPER_DOMS_privileged        (1UL<<XEN_HYPER__DOMS_privileged)
+ /* Is this guest being debugged by dom0? */
+#define XEN_HYPER__DOMS_debugging        2
+#define XEN_HYPER_DOMS_debugging         (1UL<<XEN_HYPER__DOMS_debugging)
+ /* Are any VCPUs polling event channels (SCHEDOP_poll)? */
+#define XEN_HYPER__DOMS_polling          3
+#define XEN_HYPER_DOMS_polling           (1UL<<XEN_HYPER__DOMS_polling)
+ /* Domain is paused by controller software? */
+#define XEN_HYPER__DOMS_ctrl_pause       4
+#define XEN_HYPER_DOMS_ctrl_pause        (1UL<<XEN_HYPER__DOMS_ctrl_pause)
+ /* Is this guest dying (i.e., a zombie)? */
+#define XEN_HYPER__DOMS_dying            5
+#define XEN_HYPER_DOMS_dying             (1UL<<XEN_HYPER__DOMS_dying)
+ /* In process of shutting down? */
+#define XEN_HYPER__DOMS_shuttingdown     6
+#define XEN_HYPER_DOMS_shuttingdown      (1UL<<XEN_HYPER__DOMS_shutdown)
+ /* Fully shut down? */
+#define XEN_HYPER__DOMS_shutdown         7
+#define XEN_HYPER_DOMS_shutdown          (1UL<<XEN_HYPER__DOMS_shutdown)
+
 /*
  * VCPU
  */
@@ -284,8 +320,9 @@ typedef uint32_t	Elf_Word;
 #define XEN_HYPER_ELF_NOTE_FILL_T_NOTE		(0)
 #define XEN_HYPER_ELF_NOTE_FILL_T_CORE		(1)
 #define XEN_HYPER_ELF_NOTE_FILL_T_XEN_CORE	(2)
-#define XEN_HYPER_ELF_NOTE_FILL_T_PRS		(3)
-#define XEN_HYPER_ELF_NOTE_FILL_T_XEN_REGS	(4)
+#define XEN_HYPER_ELF_NOTE_FILL_T_XEN_CORE_M	(3)
+#define XEN_HYPER_ELF_NOTE_FILL_T_PRS		(4)
+#define XEN_HYPER_ELF_NOTE_FILL_T_XEN_REGS	(5)
 
 /*
  * Command interface structs
@@ -339,6 +376,9 @@ struct xen_hyper_cmd_args {
  */
 struct xen_hyper_machdep_table {
 	void (*pcpu_init)(void);
+#ifdef IA64
+	long frame_table;
+#endif
 };
 
 struct xen_hyper_table {
@@ -385,11 +425,12 @@ struct xen_hyper_dumpinfo_context_xen_info {
 struct xen_hyper_dumpinfo_table {
 	uint note_ver;
 	struct xen_hyper_dumpinfo_context *context_array;
-	struct xen_hyper_dumpinfo_context_xen_core context_xen_core;
+	struct xen_hyper_dumpinfo_context_xen_core *context_xen_core_array;
 	struct xen_hyper_dumpinfo_context_xen_info context_xen_info;
 	char *crash_note_core_array;
-	char *crash_note_xen_core_ptr;
+	char *crash_note_xen_core_array;
 	char *crash_note_xen_info_ptr;
+	uint xen_info_cpu;
 	Elf_Word note_size;
 	Elf_Word core_offset;
 	Elf_Word core_size;
@@ -512,7 +553,7 @@ struct xen_hyper_symbol_table_data {
 };
 
 struct xen_hyper_size_table {
-	long ELF_Prstatus;			/* elf note v1, v2, v3 */
+	long ELF_Prstatus;			/* elf note v1,v2,v3,v4 */
 	long ELF_Signifo;
 	long ELF_Gregset;
 	long ELF_Timeval;
@@ -529,8 +570,8 @@ struct xen_hyper_size_table {
 	long crash_note_xen_t;			/* elf note v2 */
 	long crash_note_xen_core_t;		/* elf note v3 */
 	long crash_note_xen_info_t;		/* elf note v3 */
-	long crash_xen_core_t;			/* elf note v3 */
-	long crash_xen_info_t;			/* elf note v3 */
+	long crash_xen_core_t;			/* elf note v3,v4 */
+	long crash_xen_info_t;			/* elf note v3,v4 */
 	long domain;
 #ifdef IA64
 	long mm_struct;
@@ -614,6 +655,14 @@ struct xen_hyper_offset_table {
 	long domain_next_in_list;
 	long domain_domain_flags;
 	long domain_evtchn;
+	long domain_is_hvm;
+	long domain_is_privileged;
+	long domain_debugger_attached;
+	long domain_is_polling;
+	long domain_is_dying;
+	long domain_is_paused_by_controller;
+	long domain_is_shutting_down;
+	long domain_is_shut_down;
 	long domain_vcpu;
 	long domain_arch;
 #ifdef IA64
